require 'sequel'
require 'securerandom'

module Resta

  # Mixins for subclass for `Sequel::Model`.
  module Model

    # Current class isn't inherited from Sequel::Model.
    class UnsupportedClassError < TypeError
    end

    # Some necessary field missing for current database.
    # e.g.`_$version` field used for caching
    class NecessaryFieldMissingError < StandardError
    end

    # Class methods for model after been included.
    module ClassMethods
      attr_reader :use_increased_id, :timestamps_opts

      def enable_increased_id
        @use_increased_id = true
      end

      def use_timestamps(fields_opts=nil)
        fields_opts ||= { update_on_create: true }
        self.plugin :timestamps, fields_opts
      end
    end

    # Instance methods for model after been included.
    module InstanceMethods
      def before_create
        return false if super == false
        update_version
        # generate a v4 random UUID for resource id unless use auto increase.
        self.id ||= SecureRandom.uuid unless self.class.use_increased_id
      end

      def before_update
        return false if super == false
        update_version
      end

      private
      def update_version
        self[:'_$version'] = SecureRandom.hex(8)
      end
    end

    COMMON_DATABASE_FIELDS = {:'_$version' => :string}.freeze

    def self.included(receiver)
      raise UnsupportedClassError.new(
        "#{receiver.class.name} is not a subclass of Sequel::Model"
      ) unless receiver < Sequel::Model
      necessary_fields_checker receiver
      receiver.extend ClassMethods
      receiver.send :include, InstanceMethods
    end

    def self.necessary_fields_checker(receiver)
      COMMON_DATABASE_FIELDS.each_pair do |field, type|
        raise NecessaryFieldMissingError.new(
          "Missing field `#{field}`##{type} in #{receiver.table_name}"
        ) unless(receiver.db_schema.has_key?(field) &&
                 receiver.db_schema[field][:type] == type)
      end
    end

    private_class_method :necessary_fields_checker
  end

  # Sequel migration helpers for resta.
  module MigrationHelpers

    DEFAULT_PRIMARY_COLUMN_OPTS = {
      type: :string,
      auto_increment: false,
    }.freeze

    # Helpers used in schema modification.
    module AlterTable
      def add_primary_column(name, opts=nil)
        opts = (opts || {}).merge DEFAULT_PRIMARY_COLUMN_OPTS
        add_primary_key name, opts
      end

      def add_timestamp_columns
        [:created_at, :updated_at].each do
          |f| add_column f, :datetime
        end
      end

      def add_version_column
        add_column :'_$version', :string
      end
    end

    # Helpers for AlterTableGenerator
    module Reverser

      private

      def add_primary_column(name, opts=nil)
        opts = (opts || {}).merge DEFAULT_PRIMARY_COLUMN_OPTS
        @actions << [:drop_column, name, opts]
      end

      def add_timestamp_columns
        @actions << [:drop_column, :created_at]
        @actions << [:drop_column, :updated_at]
      end

      def add_version_column
        @actions << [:drop_column, :'_$version']
      end
    end

    # Helpers used in create table
    module CreateTable
      def primary_column(name, opts=nil)
        opts = (opts || {}).merge DEFAULT_PRIMARY_COLUMN_OPTS
        primary_key name, :string, opts
      end

      def timestamp_columns
        [:created_at, :updated_at].each do
          |f| column f, :datetime
        end
      end

      def version_column
        column :'_$version', :string
      end
    end
  end

end

module Sequel

  class MigrationAlterTableReverser
    include Resta::MigrationHelpers::Reverser
  end

  module Schema
    class AlterTableGenerator
      include Resta::MigrationHelpers::AlterTable
    end

    class CreateTableGenerator
      include Resta::MigrationHelpers::CreateTable
    end
  end
end
